# plots.py
#
# Plotting code for Variant Filtration with Neural Nets
# This includes evaluation plots like Precision and Recall curves,
# various flavors of Reciever Operating Characteristic (ROC curves),
# As well as graphs of the metrics that are watched during neural net training.
#
# December 2016
# Sam Friedman
# sam@broadinstitute.org

# Python 2/3 friendly
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import

# Imports
import os
import sys
import math
import pickle
import matplotlib
matplotlib.use('Agg') # Need this to write images from the GSA servers.  Order matters:
import matplotlib.pyplot as plt # First import matplotlib, then use Agg, then import plt

import numpy as np
from sklearn.metrics import roc_curve, auc, roc_auc_score, precision_recall_curve, average_precision_score

image_ext = '.png'

color_array = ['red', 'indigo', 'cyan', 'pink', 'purple']
key_colors = {
    'Neural Net':'green', 'CNN_SCORE':'green', 'CNN_2D':'green',
    'Heng Li Hard Filters':'lightblue',
    'GATK Hard Filters':'orange','GATK Signed Distance':'darksalmon',
    'VQSR gnomAD':'cornflowerblue', 'VQSR Single Sample':'blue', 'VQSLOD':'cornflowerblue',
    'Deep Variant':'magenta', 'QUAL':'magenta', 'DEEP_VARIANT_QUAL':'magenta',
    'Random Forest':'darkorange',
    'SNP':'cornflowerblue', 'NOT_SNP':'orange', 'INDEL':'green', 'NOT_INDEL':'red',
    'VQSLOD none':'cornflowerblue', 'VQSLOD strModel':'orange', 'VQSLOD default':'green',
    'REFERENCE':'green', 'HET_SNP':'cornflowerblue', 'HOM_SNP':'blue', 'HET_DELETION':'magenta',
    'HOM_DELETION':'violet', 'HET_INSERTION':'orange', 'HOM_INSERTION':'darkorange'
}

precision_label = 'Precision | Positive Predictive Value | TP/(TP+FP)'
recall_label = 'Recall | Sensitivity | True Positive Rate | TP/(TP+FN)'
fallout_label = 'Fallout | 1 - Specificity | False Positive Rate | FP/(FP+TN)'


def plot_precision_recall_from_scores(truth, scores, title):
    # Compute Precision-Recall and plot curve
    precision = dict()
    recall = dict()
    average_precision = dict()
    lw = 4.0
    plt.figure(figsize=(22,18))
    matplotlib.rcParams.update({'font.size': 32})
    for k in scores.keys():
        if k in key_colors:
            c = key_colors[k]
        else:
            c = np.random.choice(color_array)
        precision[k], recall[k], _ = precision_recall_curve(truth, scores[k])
        average_precision[k] = average_precision_score(truth, scores[k])
        plt.plot(recall[k], precision[k], lw=lw, color=c, label=k+' area = %0.3f' % average_precision[k])

    plt.ylim([-0.02, 1.03])
    plt.xlim([0.0, 1.0])

    plt.xlabel(recall_label)
    plt.ylabel(precision_label)
    matplotlib.rcParams.update({'font.size': 35})
    plt.title(format_title(title))

    plt.legend(loc="lower left")

    plot_name = "./figures/precision_recall_"+title+image_ext
    if not os.path.exists(os.path.dirname(plot_name)):
        os.makedirs(os.path.dirname(plot_name))
    plt.savefig(plot_name)
    print('Saved plot at:', plot_name)


def plot_simple_roc(truth, scores, title):
    fpr, tpr, _ = roc_curve(truth, scores)
    roc_score = roc_auc_score(truth, scores)
    plt.figure(figsize=(18,18))
    lw = 2
    plt.plot(fpr, tpr, color='darkorange',
             lw=lw, label='Area under ROC curve = %0.2f' % roc_score)
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='')
    plt.xlim([0.0, 1.0])
    plt.ylim([-0.02, 1.03])
    plt.xlabel(fallout_label)
    plt.ylabel(recall_label)
    plt.title(title.replace('_', ' ')+'\n')

    plt.legend(loc="lower right")
    plot_name = "./figures/simple_"+title+"_roc"+image_ext
    if not os.path.exists(os.path.dirname(plot_name)):
        os.makedirs(os.path.dirname(plot_name))
    plt.savefig(plot_name)
    print('Saved plot at:', plot_name)


def plot_rocs_from_scores(truth, scores, title):
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    lw = 6.0

    plt.figure(figsize=(22,18))
    matplotlib.rcParams.update({'font.size': 32})

    for k in scores.keys():
        if k in key_colors:
            c = key_colors[k]
        else:
            c = np.random.choice(color_array)

        fpr[k], tpr[k], _ = roc_curve(truth, scores[k])

        roc_auc[k] = roc_auc_score(truth, scores[k])
        plt.plot(fpr[k], tpr[k], color=c, lw=lw, label=k+' area = %0.3f' % roc_auc[k])



    plt.plot([0, 1], [0, 1], color='navy', lw=0.5, linestyle=':')
    plt.xlim([0.0, 1.0])
    plt.ylim([-0.02, 1.03])

    plt.xlabel(fallout_label)
    plt.ylabel(recall_label)

    matplotlib.rcParams.update({'font.size': 35})
    plt.title(format_title(title))

    plt.legend(loc="lower right")
    plot_name = "./figures/roc_"+title+image_ext
    if not os.path.exists(os.path.dirname(plot_name)):
        os.makedirs(os.path.dirname(plot_name))
    plt.savefig(plot_name)
    print('Saved plot at:', plot_name)


def plot_roc(model, test_data, test_truth, labels, title):
    y_pred = model.predict(test_data, batch_size=32, verbose=0)

    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()
    for i,k in enumerate(labels.keys()):
        fpr[i], tpr[i], _ = roc_curve(test_truth[:,i], y_pred[:,i])
        roc_auc[i] = auc(fpr[i], tpr[i])

    plt.figure(figsize=(22,18))
    lw = 2
    plt.plot(fpr[0], tpr[0], color='darkorange', lw=lw, label='ROC curve (area = %0.3f)' % roc_auc[0])
    plt.plot([0, 1], [0, 1], color='navy', lw=lw, linestyle='--')
    plt.xlim([0.0, 1.0])
    plt.ylim([-0.02, 1.03])
    plt.xlabel(fallout_label)
    plt.ylabel(recall_label)
    plt.title('ROC:' + str(labels) + '\n' + title)
    plt.legend(loc="lower right")
    plt.savefig("./figures/roc_"+title+image_ext)



def get_fpr_tpr_roc(model, test_data, test_truth, labels, batch_size=32, melt=False):
    y_pred = model.predict(test_data, batch_size=batch_size, verbose=0)

    if melt:
        melt_shape = (y_pred.shape[0]*y_pred.shape[1], y_pred.shape[2])
        y_pred = y_pred.reshape(melt_shape)
        test_truth = test_truth.reshape(melt_shape)

    return get_fpr_tpr_roc_pred(y_pred, test_truth, labels)


def get_fpr_tpr_roc_pred(y_pred, test_truth, labels):
    # Compute ROC curve and ROC area for each class
    fpr = dict()
    tpr = dict()
    roc_auc = dict()

    for k in labels.keys():
        cur_idx = labels[k]
        fpr[labels[k]], tpr[labels[k]], _ = roc_curve(test_truth[:,cur_idx], y_pred[:,cur_idx])
        roc_auc[labels[k]] = auc(fpr[labels[k]], tpr[labels[k]])

    return fpr, tpr, roc_auc


def print_auc_per_class(model, test_data, test_truth, labels):
    fpr, tpr, roc_auc = get_fpr_tpr_roc(model, test_data, test_truth, labels)
    for key in labels.keys():
        print('Label %s area under ROC: %0.3f' % (key, roc_auc[labels[key]]) )


def string_auc_per_class(model, test_data, test_truth, labels):
    string = ''
    auc_sum = 0

    fpr, tpr, roc_auc = get_fpr_tpr_roc(model, test_data, test_truth, labels)
    for key in labels.keys():
        string += '\nLabel %s area under ROC: %0.3f, AUC Area Per parameter %0.8e' % (key, roc_auc[labels[key]], (roc_auc[labels[key]]/model.count_params()))
        auc_sum += roc_auc[labels[key]]

    auc_average = auc_sum/len(labels)
    aucpp_average = auc_average/model.count_params()
    string += '\n Class Averaged AUC: %0.3f, Overall AUCPP %0.8e' % (auc_average, aucpp_average)
    return string


def get_auc(model, test_data, test_truth, labels):
    mean = 0
    fpr, tpr, roc_auc = get_fpr_tpr_roc(model, test_data, test_truth, labels)
    for key in labels.keys():
        mean += roc_auc[labels[key]]

    return mean / len(labels)


def plot_roc_per_class(model, test_data, test_truth, labels, title, batch_size=32, prefix='./figures/', melt=False):
    fpr, tpr, roc_auc = get_fpr_tpr_roc(model, test_data, test_truth, labels, batch_size, melt)

    lw = 3
    plt.figure(figsize=(28,22))
    matplotlib.rcParams.update({'font.size': 34})


    for key in labels.keys():
        if key in key_colors:
            color = key_colors[key]
        else:
            color = np.random.choice(color_array)
        plt.plot( fpr[labels[key]], tpr[labels[key]], color=color, lw=lw, label=str(key)+' area under ROC: %0.3f'%roc_auc[labels[key]]  )


    plt.plot([0, 1], [0, 1], 'k:', lw=0.5)
    plt.xlim([0.0, 1.0])
    plt.ylim([-0.02, 1.03])
    plt.xlabel(fallout_label)
    plt.ylabel(recall_label)
    plt.title('ROC:'+ title + '\n')

    matplotlib.rcParams.update({'font.size': 56})
    plt.legend(loc="lower right")
    figure_path = prefix+"per_class_roc_"+title+image_ext
    if not os.path.exists(os.path.dirname(figure_path)):
        os.makedirs(os.path.dirname(figure_path))
    plt.savefig(figure_path)
    print('Saved figure at:', figure_path)


def plot_roc_per_class_predictions(predictions, test_truth, labels, title, prefix='./figures/'):
    fpr, tpr, roc_auc = get_fpr_tpr_roc_pred(predictions, test_truth, labels)
    lw = 3
    plt.figure(figsize=(28,22))
    matplotlib.rcParams.update({'font.size': 34})

    for key in labels.keys():
        if key in key_colors:
            color = key_colors[key]
        else:
            color = np.random.choice(color_array)
        plt.plot( fpr[labels[key]], tpr[labels[key]], color=color, lw=lw, label=str(key)+' area under ROC: %0.3f'%roc_auc[labels[key]]  )

    plt.plot([0, 1], [0, 1], 'k:', lw=0.5)
    plt.xlim([0.0, 1.0])
    plt.ylim([-0.02, 1.03])
    plt.xlabel(fallout_label)
    plt.ylabel(recall_label)
    plt.title('ROC:'+ title + '\n')

    matplotlib.rcParams.update({'font.size': 56})
    plt.legend(loc="lower right")
    figure_path = prefix+"per_class_roc_"+title+image_ext
    if not os.path.exists(os.path.dirname(figure_path)):
        os.makedirs(os.path.dirname(figure_path))
    plt.savefig(figure_path)
    print('Saved figure at:', figure_path)


def plot_precision_recall_per_class_predictions(predictions, truth, labels, title, prefix='./figures/'):
    # Compute Precision-Recall and plot curve
    precision = dict()
    recall = dict()
    average_precision = dict()
    lw = 4.0
    plt.figure(figsize=(22,18))
    matplotlib.rcParams.update({'font.size': 34})

    for k in labels.keys():
        if k in key_colors:
            c = key_colors[k]
        else:
            c = np.random.choice(color_array)

        precision[k], recall[k], _ = precision_recall_curve(truth[:, labels[k]], predictions[:, labels[k]])
        average_precision[k] = average_precision_score(truth[:, labels[k]], predictions[:, labels[k]])
        plt.plot(recall[k], precision[k], lw=lw, color=c, label=k+' area = %0.3f' % average_precision[k])

    plt.ylim([-0.02, 1.03])
    plt.xlim([0.0, 1.00])

    plt.xlabel(recall_label)
    plt.ylabel(precision_label)
    plt.title(title)

    plt.legend(loc="lower left")

    plot_name = prefix+"precision_recall_"+title+image_ext
    if not os.path.exists(os.path.dirname(plot_name)):
        os.makedirs(os.path.dirname(plot_name))
    plt.savefig(plot_name)
    print('Saved plot at:%s' % plot_name)


def plot_metric_history(history, title, prefix='./figures/'):
    # list all data in history
    print(history.history.keys())
    num_plots = len ([k for k in history.history.keys() if not 'val' in k])

    row = 0
    col = 0
    rows = 4
    cols = max(2, int(math.ceil(num_plots/float(rows))))

    f, axes = plt.subplots(rows, cols, sharex=True, figsize=(36, 24))
    for k in sorted(history.history.keys()):
        if 'val' not in k:
            axes[row, col].plot(history.history[k])
            axes[row, col].set_ylabel(str(k))
            axes[row, col].set_xlabel('epoch')
            if 'val_'+k in history.history:
                axes[row, col].plot(history.history['val_'+k])
                labels = ['train', 'valid']
            else:
                labels = [k]
            axes[row, col].legend(labels, loc='upper left')

            row += 1
            if row == rows:
                row = 0
                col += 1
                if row*col >= rows*cols:
                    break

    axes[0, 1].set_title(title)
    figure_path = prefix+"metric_history_"+title+image_ext
    if not os.path.exists(os.path.dirname(figure_path)):
        os.makedirs(os.path.dirname(figure_path))
    plt.savefig(figure_path)


def read_tensor_to_image(args, tensor):
    eps = 1e-4
    bottom = np.amin(tensor)
    top = np.amax(tensor)

    print('Tensor bottom and top:', bottom, top , tensor.shape)

    tensor -= bottom
    tensor /= (eps+(top-bottom))
    tensor *= 254.0
    tensor = tensor.astype(np.uint8)

    if not args.channels_last:
        tensor = tensor.transpose((1,2,0))
        print('Transposed Tensor shape is:', tensor.shape)

    plt.imshow(tensor[:,:,:3])
    plt.show()


def format_title(title):
    t = title.split('/')[-1].replace('_', ' ')
    tsplit = t.split('true')
    if len(tsplit) > 1:
        t = tsplit[0]
        t += '\ntrue' + tsplit[-1]
    return t


def weight_path_to_title(wp):
    return wp.split('/')[-1].replace('__', '-').split('.')[0]




# Imports
import os
import sys
import h5py
import time
import numpy as np
from collections import Counter, defaultdict, namedtuple

# Package Imports
from . import defines
from . import tensor_maps

# Keras Imports
import keras.backend as K


READ_ELEMENTS = 8
Read = namedtuple("Read", "seq qual cigar reverse mate_reverse first mapping_quality reference_start")
Variant = namedtuple("Variant", "contig pos ref alt type")

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# ~~~~~~~ Inference ~~~~~~~~~~~~~~~
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
def score_and_write_batch(args, model, file_out, fifo, batch_size, python_batch_size):
	'''Score a batch of variants with a CNN model. Write tab delimited temp file with scores.

	This function is tightly coupled with the CNNVariantScore.java
	It requires data written to the fifo in the order given by transferToPythonViaFifo

	Arguments
		args: Namespace with command line or configuration file set arguments
		model: a keras model
		file_out: The VCF file where variants scores are written
		fifo: The fifo opened by GATK Streaming executor
		batch_size: The total number of variants available in the fifo
		python_batch_size: the number of variants to process in each inference
		use_reads : Use read data from a BAM for 2D CNNs or just reference for 1D CNNs
	'''
	annotation_batch = []
	reference_batch = []
	variant_types = []
	variant_data = []
	read_batch = []

	for i in range(batch_size):
		fifo_line = fifo.readline()
		fifo_data = fifo_line.split(defines.SEPARATOR_CHAR)

		variant_data.append(fifo_data[0] + '\t' + fifo_data[1] + '\t' + fifo_data[2] + '\t' + fifo_data[3])
		reference_batch.append(reference_string_to_tensor(fifo_data[4]))
		annotation_batch.append(annotation_string_to_tensor(args, fifo_data[5]))
		variant_types.append(fifo_data[6])

		fidx = 7 # 7 Because above we parsed: contig pos ref alt reference_string annotation variant_type
		if args.tensor_name in defines.TENSOR_MAPS_2D and len(fifo_data) > fidx:
			read_tuples = []
			var = Variant(fifo_data[0], int(fifo_data[1]), fifo_data[2], fifo_data[3], fifo_data[6])
			while fidx+7 < len(fifo_data):
				read_tuples.append( Read(fifo_data[fidx],
										 [int(q) for q in fifo_data[fidx+1].split(',')],
										 fifo_data[fidx+2],
										 bool_from_java(fifo_data[fidx+3]),
										 bool_from_java(fifo_data[fidx+4]),
										 bool_from_java(fifo_data[fidx+5]),
										 int(fifo_data[fidx+6]),
										 int(fifo_data[fidx+7])))
				fidx += READ_ELEMENTS
			_, ref_start, _ = get_variant_window(args, var)
			insert_dict = get_inserts(args, read_tuples, var)
			read_batch.append(read_tuples_to_read_tensor(args, read_tuples, ref_start, insert_dict))

	if args.tensor_name in defines.TENSOR_MAPS_1D:
		predictions = model.predict([np.array(reference_batch), np.array(annotation_batch)], batch_size=python_batch_size)
	elif args.tensor_name in defines.TENSOR_MAPS_2D:
		if len(read_batch) > 0:
			predictions = model.predict({'read_tensor':np.array(read_batch), 'annotations':np.array(annotation_batch)}, batch_size=python_batch_size)
	else:
		raise ValueError('Unknown tensor mapping.  Check architecture file.', args.tensor_name)

	indel_scores = predictions_to_indel_scores(predictions)
	snp_scores = predictions_to_snp_scores(predictions)

	for i in range(batch_size):
		if 'SNP' == variant_types[i]:
			file_out.write(variant_data[i]+'\t{0:.3f}'.format(snp_scores[i])+'\n')
		elif 'INDEL' == variant_types[i]:
			file_out.write(variant_data[i]+'\t{0:.3f}'.format(indel_scores[i])+'\n')
		else:
			file_out.write(variant_data[i]+'\t{0:.3f}'.format(snp_scores[i])+'\n')


def reference_string_to_tensor(reference):
	dna_data = np.zeros((len(reference), len(defines.DNA_SYMBOLS)))
	for i,b in enumerate(reference):
		if b in defines.DNA_SYMBOLS:
			dna_data[i, defines.DNA_SYMBOLS[b]] = 1.0
		elif b in defines.AMBIGUITY_CODES:
			dna_data[i] = defines.AMBIGUITY_CODES[b]
		else:
			raise ValueError('Error! Unknown code:', b)

	return dna_data


def annotation_string_to_tensor(args, annotation_string):
	name_val_pairs = annotation_string.split(';')
	name_val_arrays = [p.split('=') for p in name_val_pairs]
	annotation_map = {str(p[0]).strip() : p[1] for p in name_val_arrays if len(p) > 1}
	annotation_data = np.zeros(( len(defines.ANNOTATIONS[args.annotation_set]),))
	
	for i,a in enumerate(defines.ANNOTATIONS[args.annotation_set]):
		if a in annotation_map:
			annotation_data[i] = annotation_map[a]
	
	return annotation_data


def get_inserts(args, read_tuples, variant, sort_by='base'):
	'''A dictionary mapping insertions to reference positions.

	Ignores artificial haplotype read group.
	Relies on pysam's cigartuples structure see: http://pysam.readthedocs.io/en/latest/api.html
	Match, M -> 0
	Insert, I -> 1
	Deletion, D -> 2
	Ref Skip, N -> 3
	Soft Clip, S -> 4

	Arguments:
		args.read_limit: maximum number of reads to return
		samfile: the BAM (or BAMout) file
		variant: the variant around which reads will load

	Returns:
		insert_dict: a dict mapping read indices to max insertions at that point
	'''
	insert_dict = {}

	idx_offset, ref_start, ref_end = get_variant_window(args, variant)

	for read in read_tuples:
		index_dif = ref_start - read.reference_start
		if abs(index_dif) >= args.window_size:
			continue

		if 'I' in read.cigar:
			cur_idx = 0
			for t in cigar_string_to_tuples(read.cigar):
				if t[0] == defines.CIGAR_CODE['I']:
					insert_idx = cur_idx - index_dif
					if insert_idx not in insert_dict:
						insert_dict[insert_idx] = t[1]
					elif insert_dict[insert_idx] < t[1]:
						insert_dict[insert_idx] = t[1]

				if t[0] in [defines.CIGAR_CODE['M'], defines.CIGAR_CODE['I'], defines.CIGAR_CODE['S'], defines.CIGAR_CODE['D']]:
					cur_idx += t[1]


	read_tuples.sort(key=lambda read: read.reference_start)
	if sort_by == 'base':
		read_tuples.sort(key=lambda read: get_base_to_sort_by(read, variant))

	return insert_dict


def get_base_to_sort_by(read, variant):
	if len(read.seq) > 0:
		max_idx = len(read.seq)-1
	else:
		return 'Z'

	if variant.type != 'INDEL':
		return read.seq[clamp((variant.pos-read.reference_start)-1, 0, max_idx)]
	else:
		var_idx = variant.pos-read.reference_start
		cur_idx = 0
		for cur_op, length in cigar_string_to_tuples(read.cigar):
			cur_idx += length
			if cur_idx > var_idx:
				if cur_op == defines.CIGAR_CODE['M']:
					return read.seq[clamp(var_idx, 0, max_idx)]
				else:
					return defines.CODE2CIGAR[cur_op]
		return 'Y'


def cigar_string_to_tuples(cigar):
	if not cigar or len(cigar) == 0:
		return []
	parts = defines.CIGAR_REGEX.findall(cigar)
	# reverse order
	return [(defines.CIGAR2CODE[y], int(x)) for x,y in parts]


def get_variant_window(args, variant):
	index_offset = (args.window_size//2)
	reference_start = variant.pos-(index_offset+1)
	reference_end = variant.pos+index_offset
	return index_offset, reference_start, reference_end

def bool_from_java(val):
	return val == 'true'

def clamp(n, minn, maxn):
	return max(min(maxn, n), minn)


def read_tuples_to_read_tensor(args, read_tuples, ref_start, insert_dict):
	'''Create a read tensor based on a tensor channel map.

    Assumes read pairs have the same name.
    Only loads reads that might align inside the tensor.

    Arguments:
        args.read_limit: maximum number of reads to return
        good_reads: list of reads to make arrays from
        ref_start: the beginning of the window in reference coordinates
        insert_dict: a dict mapping read indices to max insertions at that point.

    Returns:
        tensor: 3D read tensor.
    '''
	channel_map = tensor_maps.get_tensor_channel_map_from_args(args)
	tensor = np.zeros( tensor_maps.tensor_shape_from_args(args) )

	for j,read in enumerate(read_tuples):
		rseq, rqual = sequence_and_qualities_from_read(args, read, ref_start, insert_dict)
		flag_start = -1
		flag_end = 0

		for i,b in enumerate(rseq):

			if i == args.window_size:
				break
			if b == defines.SKIP_CHAR:
				continue
			elif flag_start == -1:
				flag_start = i
			else:
				flag_end = i

			if b in args.input_symbols:
				if b == defines.INDEL_CHAR:
					if K.image_data_format() == 'channels_last':
						tensor[j, i, args.input_symbols[b]] = 1.0
					else:
						tensor[args.input_symbols[b], j, i] = 1.0
				else:
					hot_array = quality_from_mode(args, rqual[i], b, args.input_symbols)
					if K.image_data_format() == 'channels_last':
						tensor[j, i, :4] = hot_array
					else:
						tensor[:4, j, i] = hot_array
			elif b in defines.AMBIGUITY_CODES:
				if K.image_data_format() == 'channels_last':
					tensor[j, i, :4] = defines.AMBIGUITY_CODES[b]
				else:
					tensor[:4, j, i] = defines.AMBIGUITY_CODES[b]
			else:
				raise ValueError('Unknown symbol in seq block:', b)

		if K.image_data_format() == 'channels_last':
			tensor[j, flag_start:flag_end, channel_map['flag_bit_4']] = 1.0 if read.reverse else 0.0
			tensor[j, flag_start:flag_end, channel_map['flag_bit_5']] = 1.0 if read.mate_reverse else 0.0
			tensor[j, flag_start:flag_end, channel_map['flag_bit_6']] = 1.0 if read.first else 0.0
			tensor[j, flag_start:flag_end, channel_map['flag_bit_7']] = 0.0 if read.first else 1.0
		else:
			tensor[channel_map['flag_bit_4'], j,  flag_start:flag_end] = 1.0 if read.reverse else 0.0
			tensor[channel_map['flag_bit_5'], j,  flag_start:flag_end] = 1.0 if read.mate_reverse else 0.0
			tensor[channel_map['flag_bit_6'], j, flag_start:flag_end] = 1.0 if read.first else 0.0
			tensor[channel_map['flag_bit_7'], j, flag_start:flag_end] = 0.0 if read.first else 1.0

		if 'mapping_quality' in channel_map:
			if K.image_data_format() == 'channels_last':
				tensor[j, flag_start:flag_end, channel_map['mapping_quality']] = float(read.mapping_quality) / defines.MAPPING_QUALITY_MAX
			else:
				tensor[channel_map['mapping_quality'], j, flag_start:flag_end] = float(read.mapping_quality) / defines.MAPPING_QUALITY_MAX
	return tensor


def sequence_and_qualities_from_read(args, read, ref_start, insert_dict):
	cur_idx = 0
	my_indel_dict = {}
	no_qual_filler = 0

	index_dif = ref_start - read.reference_start
	for t in cigar_string_to_tuples(read.cigar):
		my_ref_idx = cur_idx - index_dif
		if t[0] == defines.CIGAR_CODE['I'] and my_ref_idx in insert_dict:
			my_indel_dict[my_ref_idx] = insert_dict[my_ref_idx] - t[1]
		elif t[0] == defines.CIGAR_CODE['D']:
			my_indel_dict[my_ref_idx] = t[1]
		if t[0] in [defines.CIGAR_CODE['M'], defines.CIGAR_CODE['I'], defines.CIGAR_CODE['S'], defines.CIGAR_CODE['D']]:
			cur_idx += t[1]

	for k in insert_dict.keys():
		if k not in my_indel_dict:
			my_indel_dict[k] = insert_dict[k]

	rseq = read.seq[:args.window_size]
	rqual = read.qual[:args.window_size]

	if index_dif > 0:
		rseq = rseq[index_dif:]
		rqual = rqual[index_dif:]
	elif index_dif < 0:
		rseq = defines.SKIP_CHAR * (-index_dif) + rseq
		rqual = [no_qual_filler]*(-index_dif) + rqual

	for j in sorted(my_indel_dict.keys(), key=int, reverse=True):
		if j < 1:
			rseq = (defines.INDEL_CHAR * my_indel_dict[j]) + rseq
			rqual = ([no_qual_filler]*my_indel_dict[j]) + rqual
		else:
			rseq = rseq[:j] + (defines.INDEL_CHAR * my_indel_dict[j]) + rseq[j:]
			rqual = rqual[:j] + ([no_qual_filler]*my_indel_dict[j]) + rqual[j:]

	return rseq, rqual


def base_quality_to_phred_array(base_quality, base, base_dict):
	phred = np.zeros((4,))
	exponent = float(-base_quality) / 10.0
	p = 1.0-(10.0**exponent) # Convert to probability
	not_p = (1.0-p) / 3.0 # Error could be any of the other 3 bases
	not_base_quality = -10 * np.log10(not_p) # Back to Phred

	for b in base_dict.keys():
		if b == defines.INDEL_CHAR:
			continue
		elif b == base:
			phred[base_dict[b]] = base_quality
		else:
			phred[base_dict[b]] = not_base_quality
	return phred


def base_quality_to_p_hot_array(base_quality, base, base_dict):
	phot = np.zeros((4,))
	exponent = float(-base_quality) / 10.0
	p = 1.0-(10.0**exponent)
	not_p = (1.0-p)/3.0

	for b in base_dict.keys():
		if b == base:
			phot[base_dict[b]] = p
		elif b == defines.INDEL_CHAR:
			continue
		else:
			phot[base_dict[b]] = not_p

	return phot


def quality_from_mode(args, base_quality, base, base_dict):
	if args.base_quality_mode == 'phot':
		return base_quality_to_p_hot_array(base_quality, base, base_dict)
	elif args.base_quality_mode == 'phred':
		return base_quality_to_phred_array(base_quality, base, base_dict)
	elif args.base_quality_mode == '1hot':
		one_hot = np.zeros((4,))
		one_hot[base_dict[base]] = 1.0
		return one_hot
	else:
		raise ValueError('Unknown base quality mode:', args.base_quality_mode)


def predictions_to_snp_scores(predictions, eps=1e-7):
	snp = predictions[:, defines.SNP_INDEL_LABELS['SNP']]
	not_snp = predictions[:, defines.SNP_INDEL_LABELS['NOT_SNP']]
	return np.log(eps + snp / (not_snp + eps))


def predictions_to_indel_scores(predictions, eps=1e-7):
	indel = predictions[:, defines.SNP_INDEL_LABELS['INDEL']]
	not_indel = predictions[:, defines.SNP_INDEL_LABELS['NOT_INDEL']]
	return np.log(eps + indel / (not_indel + eps))


def predictions_to_snp_indel_scores(predictions):
	snp_dict = predictions_to_snp_scores(predictions)
	indel_dict = predictions_to_indel_scores(predictions)
	return snp_dict, indel_dict


